"""API for eccenca DataPlatform /api/admin/backup/*."""

from requests_toolbelt.multipart import encoder
from cmem.cmempy import config


from cmem.cmempy.api import request


def get_zip():
    """Get a full backup of the knowledge graphs."""
    uri = config.get_dp_api_endpoint() + "/api/admin/backup/zip"
    return request(uri, method="GET", stream=True)


def post_zip(zip_file):
    """Post a full backup of the knowledge graphs."""
    uri = config.get_dp_api_endpoint() + "/api/admin/restore/zip"
    with open(zip_file, "rb") as file:
        files = [("file", ("backup.zip", file, "application/zip"))]
        # https://stackoverflow.com/questions/42281387/
        multipart_encoder = encoder.MultipartEncoder(
            fields=files,
            boundary="xxx---------------xxx",
        )
        return request(
            uri,
            method="POST",
            data=multipart_encoder,
            headers={"Content-Type": multipart_encoder.content_type},
            stream=True,
        )



"""API methods for working with linking tasks."""
from cmem.cmempy import config

from cmem.cmempy.api import send_request


def get_task_uri():
    """Get endpoint URI pattern for a linking task."""
    return config.get_di_api_endpoint() + "/linking/tasks/{}/{}"


def get_rule_uri():
    """Get endpoint URI pattern for a linking task rule."""
    return config.get_di_api_endpoint() + "/linking/tasks/{}/{}/rule"


def get_linking_task(project_name, task_name):
    """GET retrieve linking task."""
    response = send_request(
        get_task_uri().format(project_name, task_name), method="GET"
    )
    return response.decode("utf-8")


# pylint: disable-msg=too-many-arguments
def make_new_linking_task(
    project_name,
    task_name,
    source,
    target,
    output,
    source_type=None,
    target_type=None,
    source_restriction=None,
    target_restriction=None,
):
    """Create empty linking task."""
    params = {
        "source": source,
        "target": target,
        "sourceType": source_type,
        "targetType": target_type,
        "sourceRestriction": source_restriction,
        "targetRestriction": target_restriction,
        "output": output,
    }
    send_request(
        get_task_uri().format(project_name, task_name), method="PUT", params=params
    )


def delete_linking_task(project_name, task_name):
    """DELETE remove linking task."""
    send_request(get_task_uri().format(project_name, task_name), method="DELETE")


def get_linking_rule(project_name, task_name):
    """
    GET Return linking rule.

    Linking rule format:
    <?xml version="1.0" encoding="UTF-8"?>
    <LinkageRule linkType="&lt;http://www.w3.org/2002/07/owl#sameAs&gt;">
      <Compare id="equality1" required="false" weight="1"
      metric="equality" threshold="0.0" indexing="true">
        <TransformInput id="lowerCase1" function="lowerCase">
          <Input id="sourcePath1" path="/title"/>
        </TransformInput>
        <TransformInput id="lowerCase2" function="lowerCase">
          <Input id="targetPath1" path="/title"/>
        </TransformInput>
      </Compare>
      <Filter/>
    </LinkageRule>
    """
    response = send_request(
        get_rule_uri().format(project_name, task_name), method="GET"
    )
    return response.decode("utf-8")


def create_new_linking_rule(project_name, task_name, data):
    """PUT create new linking rule."""
    headers = {"Content-Type": "application/xml"}
    send_request(
        get_rule_uri().format(project_name, task_name),
        method="PUT",
        data=data,
        headers=headers,
    )



"""API methods for working with one project."""
import json

from cmem.cmempy import config

from cmem.cmempy.api import send_request


def get_projects_uri():
    """Get endpoint URI for project list."""
    path = "/workspace/projects"
    return config.get_di_api_endpoint() + path


def get_project_uri():
    """Get endpoint URI pattern for a project."""
    path = "/workspace/projects/{}"
    return config.get_di_api_endpoint() + path


def get_projects_api_uri():
    """Get endpoint URI for the new projects API."""
    path = "/api/workspace/projects"
    return config.get_di_api_endpoint() + path


def get_project_api_uri():
    """Get endpoint URI pattern for a single project (new API)."""
    return get_projects_api_uri() + "/{}"


def get_generate_matching_link_rule_uri():
    """Get endpoint URI for generating an ontology matching linking rule"""
    path = "/ontologyMatching/generateMatchingLinkRule"
    return config.get_di_api_endpoint() + path


def get_projects():
    """GET all projects."""
    response = send_request(get_projects_uri(), method="GET")
    return json.loads(response.decode("utf-8"))


def get_project(name):
    """GET one project."""
    response = send_request(get_project_uri().format(name), method="GET")
    return json.loads(response.decode("utf-8"))


def reload_project(name):
    """Reload one project."""
    return send_request(get_project_uri().format(name) + "/reload", method="POST")


def get_prefixes(name):
    """GET prefixes of a project."""
    response = send_request(
        get_project_api_uri().format(name) + "/prefixes",
        method="GET",
        headers={"Accept": "application/json"},
    )
    return json.loads(response.decode("utf-8"))


def make_new_project(name):
    """PUT make new project."""
    response = send_request(get_project_uri().format(name), method="PUT")
    return json.loads(response.decode("utf-8"))


def make_new_project_with_metadata(id_: str, label: str = "", description: str = ""):
    """POST make new project including metadata"""
    # {"metaData":{"label":"test","description":"eee"},"id":"myid"}
    if label == "":
        label = id_
    data = {"id": id_, "metaData": {"label": label, "description": description}}
    response = send_request(
        get_projects_api_uri(),
        method="POST",
        headers={"Content-Type": "application/json"},
        data=json.dumps(data),
    )
    return json.loads(response.decode("utf-8"))


def delete_project(name):
    """DELETE remove project."""
    send_request(get_project_uri().format(name), method="DELETE")


def get_failed_tasks_report(name):
    """Get a report of tasks that could not be loaded in a project."""
    response = send_request(
        get_project_api_uri().format(name) + "/failedTasksReport",
        method="GET",
        headers={"Accept": "application/json"},
    )
    return json.loads(response.decode("utf-8"))


def create_project_with_transformation(
    transform_project_id,
    transform_task_id,
    matching_link_spec_project_id,
    matching_link_spec_id,
):
    """Generates an ontology matching linking rule in the
    specified project with the specified linking task ID"""
    data = {
        "transformProjectId": transform_project_id,
        "transformTaskId": transform_task_id,
        "matchingLinkSpecProjectId": matching_link_spec_project_id,
        "matchingLinkSpecId": matching_link_spec_id,
    }
    data = json.dumps(data).encode("utf-8")
    headers = {"Content-Type": "application/json"}
    params = None
    response = send_request(
        get_generate_matching_link_rule_uri(),
        data=data,
        headers=headers,
        method="POST",
        params=params,
    )
    return response



"""API methods for working with workspace configurations."""
from cmem.cmempy import config
from cmem.cmempy.api import get_json


def get_workspaces_uri():
    """Get the endpoint URI for a workspaces."""
    return f"{config.get_dp_api_endpoint()}/api/conf/workspaces"


def get_workspace_migrate_uri():
    """Get the endpoint URI for a workspace migration."""
    return f"{get_workspaces_uri()}/migrateConfiguration"


def migrate_workspaces():
    """Migrate workspaces."""
    return get_json(
        uri=get_workspace_migrate_uri(),
        method="POST",
    )



"""
API method to manage cmem realm clients
"""
from json import loads

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_clients_uri():
    """Get endpoint URI for client list."""
    path = f"/admin/realms/{config.get_keycloak_realm_id()}/clients"
    return config.get_keycloak_base_uri() + path


def get_client_uri():
    """Get endpoint URI pattern for a client."""
    path = f"{get_clients_uri()}/{{}}"
    return path


def get_client_secret_uri():
    """Get endpoint URI pattern for client-secret"""
    path = f"{get_client_uri()}/client-secret"
    return path


def list_clients(params=None):
    """List clients in cmem realm."""
    uri = get_clients_uri()
    list_request = send_request(uri, method="GET", params=params)
    return loads(list_request)


def list_open_id_clients(params=None):
    """List OAuth clients and with secret in cmem realm"""
    clients = list_clients(params=params)
    return [
        _ for _ in clients if _["protocol"] == "openid-connect" and "secret" in _.keys()
    ]


def get_client_by_client_id(client_id):
    """get client by client_id"""
    params = {"clientId": client_id}
    return list_clients(params=params)


def delete_client(client_id=None):
    """delete client"""
    uri = get_client_uri().format(client_id)
    send_request(uri=uri, method="DELETE")


def generate_client_secret(client_id=None):
    """Generate a new secret for the client"""
    uri = get_client_secret_uri().format(client_id)
    credentials = send_request(uri=uri, method="POST")
    return loads(credentials)


def get_client_secret(client_id=None):
    """Get the client secret for the client"""
    uri = get_client_secret_uri().format(client_id)
    credentials = send_request(uri=uri, method="GET")
    return loads(credentials)



"""API for getting tasks in the data integration workspace."""

import json

from cmem.cmempy import config

from cmem.cmempy.api import send_request


def get_task_uri(project_id=None, task_id=None):
    """Get endpoint URL for the tasks API."""
    base_path = "/workspace/projects/{}/tasks"

    if project_id and task_id:
        path = base_path + "/{}"
        return config.get_di_api_endpoint() + path.format(project_id, task_id)
    return config.get_di_api_endpoint() + base_path.format(project_id)


def get_task(project=None, task=None, with_labels=True):
    """GET a task description."""
    response = send_request(
        get_task_uri(project, task) + "?withLabels=" + str(with_labels).lower(),
        method="GET",
        headers={"Accept": "application/json"},
    )
    return json.loads(response.decode("utf-8"))


def post_task(project=None, task=None, data=None, metadata=None, task_type=None):
    """POST a task"""
    if project is None:
        raise ValueError("Project name cannot be None.")
    if task is None:
        raise ValueError("Task id cannot be None.")
    if task_type is None:
        raise ValueError("Task type cannot be None.")

    url = get_task_uri(project)
    headers = {"Content-Type": "application/json", "Accept": "application/json"}

    task_data = {"id": task, "data": data or {}}

    if metadata is not None:
        task_data["metadata"] = metadata

    task_data["data"]["taskType"] = task_type

    encoded_data = json.dumps(task_data).encode("utf-8")

    response = send_request(
        url,
        method="POST",
        data=encoded_data,
        headers=headers,
    )

    return json.loads(response.decode("utf-8"))


def delete_task(project=None, task=None):
    """DELETE a task."""
    send_request(get_task_uri(project, task), method="DELETE")


def patch_parameter(project=None, task=None, data=None):
    """PATCH a task."""
    headers = {"Content-Type": "application/json"}
    data = json.dumps(data)
    send_request(
        get_task_uri(project, task), data=data, headers=headers, method="PATCH"
    )



"""API for eccenca DataPlatform administration."""

from cmem.cmempy import config

from cmem.cmempy.api import request


def create_showcase_data(scale_factor=None):
    """Insert the showcase data.

    Args:
        scale_factor (int): multiplies the default showcase with a factor
    """
    uri = config.get_dp_api_endpoint() + "/api/admin/showcase"
    data = {}
    if scale_factor is not None:
        data["scaleFactor"] = scale_factor
    return request(uri, method="POST", data=data)


def import_bootstrap_data():
    """Replace the bootstrap data with the most current one.

    Overwrites all data in the shape, vocabulary, query and other catalogues.
    This operation overwrites data. Use with care.
    """
    uri = config.get_dp_api_endpoint() + "/api/admin/bootstrap"
    data = {}
    return request(uri, method="POST", data=data)


def delete_bootstrap_data():
    """Delete the bootstrap data."""
    uri = config.get_dp_api_endpoint() + "/api/admin/bootstrap"
    return request(uri, method="DELETE")


def get_prometheus_data():
    """Get DP /actuator/prometheus data."""
    uri = config.get_dp_api_endpoint() + "/actuator/prometheus"
    return request(uri, method="GET")



"""python package and plugin management."""
from json import loads
from os.path import basename
from requests_toolbelt.multipart import encoder

from cmem.cmempy import config
from cmem.cmempy.api import request, send_request


def install_package_by_name(package_name):
    """Install a python package by name."""
    uri = config.get_di_api_endpoint() + "/api/python/installPackageByName"
    install_request = send_request(uri, method="POST", params={"name": package_name})
    return loads(install_request)


def install_package_by_file(package_file):
    """Install a python package by source tar.gz or wheel file."""
    uri = config.get_di_api_endpoint() + "/api/python/installPackage"
    file_name = basename(package_file)
    with open(package_file, "rb") as file:
        files = [("file", (file_name, file))]
        # https://stackoverflow.com/questions/42281387/
        multipart_encoder = encoder.MultipartEncoder(
            fields=files,
            boundary="xxx---------------xxx",
        )
        install_request = request(
            uri,
            method="POST",
            data=multipart_encoder,
            headers={"Content-Type": multipart_encoder.content_type},
            stream=True,
        )
        return loads(install_request.content.decode())


def uninstall_package(package_name: str):
    """Uninstall a python package."""
    uri = config.get_di_api_endpoint() + "/api/python/uninstallPackage"
    uninstall_request = send_request(uri, method="POST", params={"name": package_name})
    return loads(uninstall_request)


def list_packages():
    """List installed python packages."""
    uri = config.get_di_api_endpoint() + "/api/python/listPackages"
    list_request = send_request(
        uri,
        method="GET",
    )
    return loads(list_request)


def update_plugins():
    """Discover plugins installed in packages."""
    uri = config.get_di_api_endpoint() + "/api/python/updatePlugins"
    update_request = send_request(
        uri,
        method="GET",
    )
    return loads(update_request)


def list_plugins():
    """List installed python packages."""
    uri = config.get_di_api_endpoint() + "/api/python/listPlugins"
    list_request = send_request(
        uri,
        method="GET",
    )
    return loads(list_request)



"""API for working with rules."""
import json

from cmem.cmempy import config

from cmem.cmempy.api import send_request


def get_transformation_rules_uri():
    """Get endpoint URI pattern for transformation task rules."""
    path = "/transform/tasks/{}/{}/rules"
    return config.get_di_api_endpoint() + path


def get_transformation_rules(project_name, task_name):
    """GET transformation rules for the task."""
    headers = {"Accept": "application/json"}
    response = send_request(
        get_transformation_rules_uri().format(project_name, task_name),
        method="GET",
        headers=headers,
    )
    return json.loads(response.decode("utf-8"))


def create_transformation_rules(project_name, task_name, data=None, xml=False):
    """GET transformation rules for the task."""
    if xml:
        headers = {"Content-Type": "application/xml"}
    else:
        headers = {"Content-Type": "application/json"}
    send_request(
        get_transformation_rules_uri().format(project_name, task_name),
        method="PUT",
        headers=headers,
        data=data,
    )



"""
API method to manage cmem realm users
"""
import json
from json import loads
from typing import Optional, Union

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_users_uri():
    """Get endpoint URI for user list."""
    path = f"/admin/realms/{config.get_keycloak_realm_id()}/users"
    return config.get_keycloak_base_uri() + path


def get_user_uri():
    """Get endpoint URI pattern for a user."""
    path = f"{get_users_uri()}/{{}}"
    return path


def get_user_group_uri():
    """Get endpoint URI pattern for a user group."""
    path = f"{get_user_groups_uri()}/{{}}"
    return path


def get_user_groups_uri():
    """Get endpoint URI pattern for a user group list."""
    path = f"{get_user_uri()}/groups"
    return path


def get_execute_actions_email_uri():
    """Get endpoint URI pattern for a execute actions email."""
    path = f"{get_user_uri()}/execute-actions-email"
    return path


def list_users(params=None):
    """List users in cmem realm."""
    uri = get_users_uri()
    list_request = send_request(uri, method="GET", params=params)
    return loads(list_request)


def request_password_change(user_id):
    """Email to update the password"""
    uri = get_execute_actions_email_uri().format(user_id)
    headers = {"Content-Type": "application/json"}
    data = ["UPDATE_PASSWORD"]
    json_data = json.dumps(data).encode("utf-8")
    send_request(uri=uri, method="PUT", headers=headers, data=json_data)


def create_user(username):
    """create user to cmem relam"""
    data = {
        "username": username,
        "email": _get_normalized_email(username),
        "enabled": True,
    }

    uri = get_users_uri()
    json_data = json.dumps(data).encode("utf-8")
    headers = {"Content-Type": "application/json"}
    send_request(uri, method="POST", headers=headers, data=json_data)


def _get_normalized_email(email: str) -> str:
    """validate and get normalized email.

    validation is done in a SIMPLE way and not sophisticated.
    in case the email is not valid, return an empty string.
    """
    email = email.lower().strip()
    if "@" not in email:
        return ""
    return email


def update_user(
    user_id: str,
    username: str,
    first_name: str,
    last_name: str,
    email: str,
    email_verified: Optional[bool] = None,
):
    """update user to cmem relam"""
    data: dict[str, Union[str, bool]] = {
        "username": username,
        "firstName": first_name,
        "lastName": last_name,
        "email": email,
    }
    if email_verified is not None:
        data["emailVerified"] = email_verified
    uri = get_user_uri().format(user_id)
    json_data = json.dumps(data).encode("utf-8")
    headers = {"Content-Type": "application/json"}
    send_request(uri, method="PUT", headers=headers, data=json_data)


def get_user_by_username(username):
    """get user by username"""
    params = {"username": username, "exact": True}
    return list_users(params=params)


def assign_groups(user_id: str, group_ids: list):
    """assign groups to a user"""
    for group_id in group_ids:
        assign_group(user_id=user_id, group_id=group_id)


def assign_group(user_id: str, group_id):
    """assign group to a user"""
    uri = get_user_group_uri().format(user_id, group_id)
    headers = {"Content-Type": "application/json"}
    send_request(uri=uri, method="PUT", headers=headers)


def unassign_groups(user_id: str, group_ids: list):
    """unassign groups to a user"""
    for group_id in group_ids:
        unassign_group(user_id=user_id, group_id=group_id)


def unassign_group(user_id: str, group_id):
    """unassign group to a user"""
    uri = get_user_group_uri().format(user_id, group_id)
    headers = {"Content-Type": "application/json"}
    send_request(uri=uri, method="DELETE", headers=headers)


def user_groups(user_id: str):
    """List groups for a user"""
    uri = get_user_groups_uri().format(user_id)
    list_request = send_request(uri=uri, method="GET")
    return loads(list_request)


def delete_user(user_id=None):
    """delete user by id"""
    uri = get_user_uri().format(user_id)
    send_request(uri=uri, method="DELETE")


def reset_password(user_id, value, temporary=False):
    """reset user password"""
    uri = get_user_uri().format(user_id) + "/reset-password"
    data = {"type": "Password", "value": value, "temporary": temporary}
    json_data = json.dumps(data).encode("utf-8")
    headers = {"Content-Type": "application/json"}
    send_request(uri=uri, method="PUT", headers=headers, data=json_data)



"""API for working with custom tasks."""
from cmem.cmempy import config

from cmem.cmempy.api import send_request


def get_custom_tasks_uri():
    """Get endpoint URI pattern for a custom task."""
    path = "/workspace/projects/{}/customTasks/{}"
    return config.get_di_api_endpoint() + path


def get_custom_task(project_name, task_name):
    """GET retrieve custom task for the project."""
    response = send_request(
        get_custom_tasks_uri().format(project_name, task_name), method="GET"
    )
    return response.decode("utf-8")


def create_custom_task(project_name, task_name, data=None):
    """PUT create custom task for the project."""
    headers = {"Content-Type": "text/xml"}
    send_request(
        get_custom_tasks_uri().format(project_name, task_name),
        method="PUT",
        data=data,
        headers=headers,
    )


def delete_custom_task(project_name, task_name):
    """DELETE remove custom task from the project."""
    send_request(
        get_custom_tasks_uri().format(project_name, task_name), method="DELETE"
    )



"""
API method to manage cmem realm groups
"""
from json import loads

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_groups_uri():
    """Get endpoint URI for group list."""
    path = f"/admin/realms/{config.get_keycloak_realm_id()}/groups"
    return config.get_keycloak_base_uri() + path


def list_groups(params=None):
    """List groups in cmem relam"""
    uri = get_groups_uri()
    list_request = send_request(uri, method="GET", params=params)
    return loads(list_request)
